using System.ComponentModel;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading.Channels;
using gatewaySCO.POSBC;
using GatewaySCO;
using Serilog;
using Serilog.Events;

namespace GatewayCHEC.Servidor;



/// <summary>
/// Constructor servidor socket. Usa ip y puerto defualt.
/// El método IniciarAync() activa el servidor.
/// <param name="direccion">Dirección IP servidor, típicamente '127.0.0.1'.</param>
/// <param name="puerto">Número de puerto para el socket, default 11.000</param>
/// <returns>Retorna tipo Task.</returns>
/// </summary>
public class ServidorGatewayCHEC(string ipGateway, int ptoGateway, string ipPOSBC, int ptoPOSBC)
{
    readonly static bool _isDebug = Log.IsEnabled(LogEventLevel.Debug);
    /// <summary>
    /// Dirección ip para vincular el socket.
    /// </summary>
    public string IpGateway { get; private set; } = ipGateway;
    /// <summary>
    /// Puerto a vincular socket.
    /// </summary>
    public Int32 PuertoGateway { get; private set; } = ptoGateway;
    /// <summary>
    /// Dirección ip servidor POSBC.
    /// </summary>
    public string IpPOSBC { get; private set; } = ipPOSBC;
    /// <summary>
    /// Puerto servidro POSBC
    /// </summary>
    public Int32 PuertoPOSBC { get; private set; } = ptoPOSBC;
    /// <summary>
    /// Longitud de la cola de conexión al socket.
    /// </summary>
    public int LongColaConexiones { get; private set; } = 10;
    /// <summary>
    /// Fija timeout usado para enviar / recibir em modo sincrónico, en milisegundos.
    /// Si el timeout se excede, se presenta una excepción. 
    /// Default a 5 segundos.
    /// </summary>
    public int TimeOutSincMs { get; private set; } = 3000;
    /// <summary>
    ///  Servidor POSBC
    /// </summary>
    static ClienteServidorPOSBC _posbc;
    long _numeroConexionesEntrantes = 0;
    static long _nroMsjEntrada = 0;
    static long _nroMsjSalida = 0;
    static Socket _socketCHEC;

    // Canal de entrada.
    private static readonly Channel<string> _canalEntrada = Channel.CreateUnbounded<string>();
    // Canal de salida.
    private static readonly Channel<string> _canalSalida = Channel.CreateUnbounded<string>();

    /// <summary>
    ///  El POSBC debe ser activado antes de invocar el método Activa.
    /// </summary>
    /// <param name="ip"></param>
    /// <param name="pto"></param>
    private void ActivaPOSBC()
    {
        _posbc = new(IpPOSBC, PuertoPOSBC);
        _posbc.AbreConexion();
    }

    public void Activa()
    {
        IPEndPoint ip = new(IPAddress.Parse(IpGateway), PuertoGateway);
        int cont = 0;
        // Clave usar 'using' para liberar correctamente recursos.
        using Socket tcpSocket = new(
            ip.AddressFamily,
            SocketType.Stream,
            ProtocolType.Tcp);

        // Configuración comportamiento socket tcp.
        // No se permite a otro socket compartir el puerto.
        // TODO - Investigar: cuando servidor y cliente están en la misma instanción de s.o.
        //          no comparten el puerto?.
        // tcpSocket.ExclusiveAddressUse = true;
        // El socket espera los segundos del parámetro para terminar de enviar
        // datos (si hay en el buffer de salida), despues que se llama Socket.Close.
        tcpSocket.LingerState = new LingerOption(true, 3);
        // Desactiva algoritmo Nagle.
        tcpSocket.NoDelay = true;
        // Timeout entrada / salida.
        tcpSocket.ReceiveTimeout = 0;
        tcpSocket.SendTimeout = TimeOutSincMs;
        //tcpSocket.Blocking = false;
        tcpSocket.Bind(ip);
        tcpSocket.Listen(LongColaConexiones);

        // Id proceso. Compilación difiere según .net usado.
        int idProceso = -1;
#if NETFRAMEWORK
                idProceso = Process.GetCurrentProcess().Id;
#elif (NETSTANDARD || NET5_0_OR_GREATER)
        idProceso = Environment.ProcessId;
#endif

        Log.Information("Gateway servidor socket en {ip} : {puerto}, proceso id {id}", IpGateway, PuertoGateway, idProceso);
        if (_isDebug)
        {
            Log.Debug("Versión framework {version} #6", Environment.Version);
            Log.Debug("Tcp Socket configuración:");
            Log.Debug($"  Blocking  {tcpSocket.Blocking}");
            Log.Debug($"  ExclusiveAddressUse {tcpSocket.ExclusiveAddressUse}");
            Log.Debug($"  LingerState {tcpSocket.LingerState.Enabled}, {tcpSocket.LingerState.LingerTime}");
            Log.Debug($"  NoDelay {tcpSocket.NoDelay}");
            Log.Debug($"  ReceiveBufferSize {tcpSocket.ReceiveBufferSize}");
            Log.Debug($"  ReceiveTimeout {tcpSocket.ReceiveTimeout}");
            Log.Debug($"  SendBufferSize {tcpSocket.SendBufferSize}");
            Log.Debug($"  SendTimeout {tcpSocket.SendTimeout}");
            Log.Debug($"  Ttl {tcpSocket.Ttl}");
            Log.Debug($"  IsBound {tcpSocket.IsBound}");
        }
        try
        {
            // Activa el POSBC
            if (_posbc == null) ActivaPOSBC();
            var tareaProcesaMsj = Task.Run(() => ProcesaEntradaCHEC());
            var tareaRecibePOSBC = Task.Run(() => _posbc.Recibe(_canalSalida));
            var tareaRemiteRespuesta = Task.Run(() => ProcesaSalidaPOSBC());

            // Procesa conexiones al Gateway.
            while (true)
            {
                Log.Information("Esperando conexión");
                Socket cliente = tcpSocket.Accept();
                Log.Information("Conexión remota ip {ip} en puerto {pto}", IPAddress.Parse(((IPEndPoint)cliente.RemoteEndPoint).Address.ToString()), ((IPEndPoint)cliente.RemoteEndPoint).Port.ToString());
                _ = Task.Run(() => AceptaEntradas(cliente));
            }
        }
        finally
        {
            tcpSocket.Close();
            _posbc.CierraConexion();
        }
    }

    public static async Task AceptaEntradas(Socket cliente)
    {
        var taskId = Task.CurrentId?.ToString() ?? "no-task";
        _socketCHEC = cliente;
        Log.Debug("{tid} Tarea acepta entradas iniciada", taskId);
        while (true)
        {
            // Lee longitud mensaje entrante, 4 bytes.
            Log.Information("{tid} Esperando CHEC", taskId);
            var bufferLongitud = new byte[4];

            // Lee los primeros 4 bytes de los datos de entrada, los cuales indican 
            // la longitud del resto del mensaje.
            int nroBytesLeidos = await cliente.ReceiveAsync(new ArraySegment<byte>(bufferLongitud), SocketFlags.None);

            // Si el número de bytes leidos es cero, la conexión se ha cerrado.
            // El método Receive es sincróno, luego si retorna sin datos, es que no hay conexión.
            // Mientras tenga conexión, se queda esperando datos.
            if (nroBytesLeidos == 0) break;

            uint longitudMensaje = Util.LongitudMensaje(bufferLongitud);
            Log.Debug("{tid} CHECK bytes cabecera {nroBytes}, longitud mensaje {longitud}", taskId, nroBytesLeidos, longitudMensaje);
            // Prepara buffer de entrada según la longitud esperada del mensaje.
            var bufferMensaje = new byte[longitudMensaje];
            //nroBytesLeidos = cliente.Receive(bufferMensaje, 0, bufferMensaje.Length, SocketFlags.None);
            nroBytesLeidos = await cliente.ReceiveAsync(new ArraySegment<byte>(bufferMensaje), SocketFlags.None);
            Log.Debug("{tid} CHECK bytes mensaje {nroBytes}", taskId, nroBytesLeidos);
            _nroMsjEntrada++;
            Log.Information("{tid} CHEC -> Gateway {nro} - {bytes} bytes cuerpo", taskId, _nroMsjEntrada, nroBytesLeidos);

            // Arreglo de bytes de entrada se transforma en string, 
            // constituye el mensaje.
            string mensaje = Encoding.UTF8.GetString(bufferMensaje, 0, nroBytesLeidos);
            Log.Verbose("{tid} CHEC -> Gateway mensaje\n{msj}", taskId, mensaje);

            // Se agrega mensaje a la cola de entrada de mensajes.
            await _canalEntrada.Writer.WriteAsync(mensaje);
            Log.Debug("{tid} mensaje -> [cola entrada]", taskId);
        }
        cliente.Close();
    }

    private static async Task ProcesaEntradaCHEC()
    {
        var taskId = Task.CurrentId?.ToString() ?? "no-task";
        Log.Debug("{tid} Tarea procesa entrada CHEC inicia", taskId);
        while (true)
        {
            // Espera por mensajes en cola.
            string mensaje = await _canalEntrada.Reader.ReadAsync();
            Log.Verbose("Procesa mensaje de [cola entrada CHEC]\n{msj}", mensaje);
            //
            //
            // PROCESA EL MENSAJE CHEC -> GATEWAY 
            //
            //
            // Remita mensaje procesado al POSBC
            await _posbc.Envia(mensaje);
        }
    }

    private static async Task ProcesaSalidaPOSBC()
    {
        var taskId = Task.CurrentId?.ToString() ?? "no-task";
        Log.Debug("{tid} Tarea procesa salida POSBC inicia", taskId);
        while (true)
        {
            string mensaje = await _canalSalida.Reader.ReadAsync();
            Log.Verbose("Procesa mensaje de [cola salida POSBC]\n{msj}", mensaje);
            //
            //
            // PROCESA MENSAJE POSBC --> CHEC
            //
            //
            byte[] bufferSalida = Util.ConvierteEnBufferBytes(mensaje);
            await _socketCHEC.SendAsync(bufferSalida);
            _nroMsjSalida++;
            Log.Information("{tid} CHEC <- Gateway {nro} - {bytes} bytes total", taskId, _nroMsjSalida, bufferSalida.Length);
            Log.Verbose("{tid} CHEC <- Gateway mensaje\n{msj}", taskId, mensaje);
        }
    }
}